import logging
import os

from jinja2 import Environment, PackageLoader

from ..client import Session
from .article import LatexArticle

logger = logging.getLogger()


class LatexProject:
    def __init__(self, session: Session, target_folder: str):
        self.session = session
        self.target_folder = os.path.abspath(target_folder)
        self.assets_folder = os.path.join(self.target_folder, "assets")
        self.images_folder = os.path.join(self.target_folder, "assets", "images")
        self.documents_folder = os.path.join(self.target_folder, "documents")
        logger.info(f"Creating {self.images_folder}")
        os.makedirs(self.images_folder, exist_ok=True)
        logger.info(f"Creating {self.documents_folder}")
        os.makedirs(self.documents_folder, exist_ok=True)
        self.articles = []
        self.reference_list = []
        self.jinja = None
        self._configure_jinja()

    @classmethod
    def build_single_article(cls,
        session: Session, target_folder: str,
        project_id: str, article_id: str, version: int):
        latex_project = cls(session, target_folder)
        try:
            latex_project.add_article(project_id, article_id, version)
        except ValueError as err:
            logging.error(f"Error: {str(err)}")
            raise ValueError(f"Could not add article to LaTeX project: {str(err)}") from err

        try:
            logging.info(f"writing to {target_folder}")
            latex_project.write()
        except ValueError as err:
            logging.error(f"Error: {str(err)}")
            logging.error(f"Could not write final document")
            raise ValueError(f"Could not write final document") from err

    def _configure_jinja(self):
        """
        Default jinja syntax doesn't play well with LaTeX.
        Create a custom environment that does.
        http://eosrei.net/articles/2015/11/latex-templates-python-and-jinja2-generate-pdfs
        """
        self.jinja = Environment(
            block_start_string=r"\BLOCK{",
            block_end_string="}",
            variable_start_string=r"\VAR{",
            variable_end_string="}",
            comment_start_string=r"\#{",
            comment_end_string="}",
            line_statement_prefix="%%",
            line_comment_prefix="%#",
            trim_blocks=True,
            autoescape=False,
            loader=PackageLoader("curvenote", "latex/templates"),
        )

    def add_article(self, project_id: str, article_id: str, version: int):
        latex_article = LatexArticle(self.session, project_id, article_id)
        latex_article.fetch(version)
        latex_article.localize(self.session, self.assets_folder, self.reference_list)

        idx = len(self.articles)
        filename = f"{idx}_{latex_article.block.name}"
        self.articles.append((f"documents/{filename}", latex_article, filename))

    def render(self):
        template = self.jinja.get_template("index.tpl")
        try:
            _, first_article, __ = self.articles[0]
            return template.render(
                article_paths=[p for p, *_ in self.articles],
                main_title=first_article.title,
                main_author_list=first_article.author_names,
                main_day=first_article.date.day,
                main_month=first_article.date.month,
                main_year=first_article.date.year,
            )
        except ValueError as err:
            raise ValueError("Need at least one article") from err

    def write(self):
        for (_, article, filename) in self.articles:
            article_filepath = os.path.join(self.documents_folder, filename + ".tex")
            article.write(article_filepath)

        with open(os.path.join(self.target_folder, "index.tex"), "w+") as file:
            file.write(self.render())

        if len(self.reference_list) > 0:
            with open(os.path.join(self.target_folder, "main.bib"), "w+") as file:
                for reference in self.reference_list:
                    file.write(reference.bibtex)



import logging
import os
import sys

import typer

from .client import Session
from .latex.project import LatexProject

app = typer.Typer()

logger = logging.getLogger()


@app.command()
def push(
    path: str = typer.Argument(..., help="Local file or folder to push to curvenote."),
    project: str = typer.Argument(
        ...,
        help=(
            "Identifier of target Project. PROJECT may match title, name, or id "
            "of an existing Project. If no existing Project is found, a new "
            "Project will be created with title PROJECT."
        ),
    ),
    article: str = typer.Option(
        None,
        help=(
            "Identifier of target Article. ARTICLE may match title, name, or id "
            "of an existing Article. If no existing Article is found, a new "
            "Article will be created with title ARTICLE. ARTICLE is ignored if "
            "PATH is a folder. If PATH is a folder or ARTICLE is not provided, "
            "filename will be used for Article."
        ),
    ),
    team: str = typer.Option(
        None,
        help=(
            "Team to use when creating a new Project. TEAM is ignored if PROJECT "
            "already exists. If PROJECT does not exist and TEAM is not provided, "
            "the new Project will be created under the current user."
        ),
    ),
    token: str = typer.Argument(
        ..., envvar="CURVENOTE_TOKEN", help="API token generated from curvenote.com"
    ),
):
    """Push contents of local file or folder to curvenote Project"""
    if not os.path.exists(path):
        raise ValueError(f"path not found: {path}")
    session = Session(token)

    typer.echo("Checking for project access")
    project_obj = session.get_or_create_project(
        title=project,
        team=team,
    )
    if os.path.isdir(path):
        typer.echo("pushing from folder")
        session.documents_from_folder(folder=path, project=project_obj)
    elif os.path.isfile(path):
        _, file_extension = os.path.splitext(path)
        if file_extension == ".ipynb":
            typer.echo("pushing notebook file...")
            session.notebook_from_file(
                filename=path, project=project_obj, title=article
            )
        elif file_extension == ".md":
            typer.echo("pushing markdown file...")
            session.article_from_file(filename=path, project=project_obj, title=article)
        else:
            raise ValueError(f"unsupported file type: {file_extension}")
    else:
        raise ValueError(f"unable to resolve path: {path}")


@app.command()
def pull_as_latex(
    target: str = typer.Argument(
        ...,
        help=(
            "Local folder in which to construct the Latex assets. If TARGET exists it"
            "and all files will be removed and a new empty folder structure created"
        ),
    ),
    project: str = typer.Argument(
        ...,
        help=(
            "Identifier of existing Project containing ARTICLE. PROJECT may match title,"
            " name, or id of an existing Project. If no existing Project is found, an "
            "error will be raised"
        ),
    ),
    article: str = typer.Argument(
        ...,
        help=(
            "Identifier of existing Article. ARTICLE may match title, name, or id "
            "of an existing Article. If no existing Article is found, an error will"
            "be raised."
        ),
    ),
    token: str = typer.Argument(
        ..., envvar="CURVENOTE_TOKEN", help="API token generated from curvenote.com"
    ),
    version: int = typer.Option(
        None,
        help=(
            "Version of the article to pull, if not specified will pull the latest version."
        ),
    ),
):
    logger.setLevel(logging.INFO)
    handler = logging.StreamHandler(sys.stdout)
    handler.setLevel(logging.INFO)
    formatter = logging.Formatter(
        "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
    )
    handler.setFormatter(formatter)
    logger.addHandler(handler)

    session = Session(token)

    try:
        project = session.get_project(project)
    except ValueError as err:
        typer.echo(f"Could not find project: {project} or you do not have access")
        raise typer.Exit(code=1) from err

    typer.echo(f"Found project: {project.name}")

    try:
        LatexProject.build_single_article(session, target, project, article, version)
    except ValueError as err:
        typer.echo(err)
        raise typer.Exit(code=1)


@app.command()
def get_me(token: str = typer.Argument(
        ..., envvar="CURVENOTE_TOKEN", help="API token generated from curvenote.com"
    )):
    session = Session(token)
    typer.echo(session.user().json(indent=4))


@app.command()
def get_my_projects(token: str = typer.Argument(
        ..., envvar="CURVENOTE_TOKEN", help="API token generated from curvenote.com"
    )):
    session = Session(token)
    for project in session.my_projects():
        typer.echo(project.json(indent=4))



import logging
from typing import List, Optional

from curvenote.latex import utils

from ..client import Session
from ..models import BlockFormat, BlockKind
from . import utils

logger = logging.getLogger()


class LatexArticle:
    """
    Class to represent an article in the latex project.
    With the abilty to fetch, localize its assets and write itself to file.
    """

    def __init__(self, session: Session, project_id: str, article_id: str):
        self.session = session
        self.project_id = project_id
        self.article_id = article_id
        self.block = None
        self.version = None
        self.children = None
        self.content = ""
        # TODO obvious one to centralise in project or someway
        self.users = []

    @property
    def title(self):
        if self.block:
            return self.block.title
        return ""

    @property
    def description(self):
        if self.block:
            return self.block.description
        return None

    @property
    def authors(self):
        if self.block:
            return self.block.authors
        return []

    @property
    def author_names(self):
        return [
            user.display_name
            for author in self.authors
            for user in self.users
            if user.id == author.user
        ]

    @property
    def date(self):
        if self.version:
            return self.version.date or self.version.date_created
        return None

    def fetch(self, version: Optional[int] = None):
        """Download article and all children in Latex format

        Raises ValueError if download fails.
        """
        block = self.session.get_block(
            self.project_id, self.article_id, kind=BlockKind.article
        )
        version_to_fetch = version or block.latest_version
        children = self.session.get_version_with_children(
            block, version_to_fetch, fmt=BlockFormat.tex
        )
        if len(children.errors) > 0:
            logger.error("There were errors fetching some children")
            for error in children.errors:
                logger.error(error)

        if children.versions.items[0].kind != BlockKind.article:
            raise ValueError("Expected first child to be an article")
        self.block = children.blocks.items[0]
        self.version, *self.children = children.versions.items
        logger.info(
            f"Processing Article: {self.version.id.project}/"
            f"{self.version.id.block}/versions/{self.version.id.version}"
        )

    def localize(
        self,
        session: Session,
        assets_folder: str,
        reference_list: List[utils.LocalReferenceItem],
    ):
        """Parse article content and pull assets to local storage

        - images
        - authors
        - citations

        TODO: this is turning into a double dispatchy thing, maybe article just holds a
        reference to the project
        """
        self._localize_authors(session)
        self._localize_content(session, assets_folder, reference_list)

    def write(self, filepath: str):
        with open(filepath, "w+") as file:
            file.write(self.content)

    def _localize_authors(self, session: Session):
        for author in self.authors:
            try:
                self.users.append(session.get_user(author.user))
            except ValueError as err:
                logger.info(f"Could note get user {author.user}: {err}")
                continue

    def _localize_content(
        self, session: Session, assets_folder: str, reference_list: List[str]
    ):
        """
        Ignores blocks of any type other than Content and Output.
        """
        concatenated_content = ""
        for child in self.children:
            if child.kind == BlockKind.content:
                logger.info("Content Block")
                content = utils.localize_images_from_content_block(
                    self.session, assets_folder, child.content
                )
                content = utils.localize_references_from_content_block(
                    session, reference_list, content
                )
            elif child.kind == BlockKind.output:
                logger.info(f"Output Block - num outputs: {len(child.outputs)}")
                content = utils.localize_images_from_output_block(assets_folder, child)
            else:
                logger.warning(f"Can't process block with kind:{child.kind} yet")
                continue

            concatenated_content += content + "\n"

        self.content = concatenated_content



from datetime import datetime
from enum import Enum
from typing import Any, Dict, Generic, List, Optional, TypeVar, Union

from pydantic import AnyHttpUrl, BaseModel
from pydantic.generics import GenericModel


class SelfLinks(BaseModel):
    self: AnyHttpUrl


class ListLinks(SelfLinks):
    next: Optional[AnyHttpUrl]


class UserLinks(SelfLinks):
    photo: AnyHttpUrl


class MyUserLinks(UserLinks):
    projects: AnyHttpUrl
    banner: Optional[AnyHttpUrl]
    access: AnyHttpUrl


class User(BaseModel):
    id: str
    username: str
    display_name: str
    bio: str
    location: str
    website: str
    github: str
    twitter: str
    affiliation: str
    orcid: str
    date_created: datetime
    links: UserLinks


class MyUser(User):
    email: str
    email_verified: bool
    links: MyUserLinks


class ProjectLinks(SelfLinks):
    access: AnyHttpUrl
    team: AnyHttpUrl
    blocks: AnyHttpUrl
    manifest: AnyHttpUrl


class ProjectVisibility(str, Enum):
    """"""

    public = "public"
    private = "private"
    manifest = "manifest"


class ProjectPost(BaseModel):
    team: str
    name: str
    title: Optional[str]
    description: Optional[str]


class Project(ProjectPost):
    id: str
    created_by: str
    visibility: ProjectVisibility
    date_created: datetime
    date_modified: datetime
    links: ProjectLinks


class ProjectList(BaseModel):
    items: List[Project]
    links: ListLinks


class Author(BaseModel):
    plain: Union[str, None]
    user: Union[str, None]


class BlockKind(str, Enum):
    """"""

    article = "Article"
    code = "Code"
    content = "Content"
    image = "Image"
    navigation = "Navigation"
    notebook = "Notebook"
    output = "Output"
    quote = "Quote"
    reference = "Reference"


class BlockID(BaseModel):
    project: str
    block: str


class BlockLinks(SelfLinks):
    project: AnyHttpUrl
    comments: Optional[AnyHttpUrl]
    versions: AnyHttpUrl
    created_by: AnyHttpUrl
    drafts: Optional[AnyHttpUrl]
    default_draft: Optional[AnyHttpUrl]
    latest: Optional[AnyHttpUrl]
    published: Optional[AnyHttpUrl]


class BlockPost(BaseModel):
    kind: BlockKind
    name: Optional[str]
    tags: Optional[List[str]]


class Block(BlockPost):
    id: BlockID
    title: str
    description: str
    latest_version: Optional[int]
    published: bool
    published_versions: Optional[List[int]]
    default_draft: Optional[str]
    num_versions: Optional[int]
    num_comments: Optional[int]
    created_by: str
    date_created: datetime
    date_modified: Optional[datetime]
    authors: Optional[List[Author]]
    links: BlockLinks


class BlockList(BaseModel):
    items: List[Block]
    links: ListLinks


class CommentID(BaseModel):
    project: str
    block: str
    comment: str


class CommentLinks(SelfLinks):
    block: AnyHttpUrl


class Comment(BaseModel):
    id: CommentID
    content: str
    created_by: str
    parent: Optional[str]
    children: List[str]
    date_created: datetime
    date_modified: datetime
    links: CommentLinks


class CommentList(BaseModel):
    items: List[Comment]
    links: ListLinks


class BlockChildID(BaseModel):
    project: str
    block: str
    version: Optional[int]
    draft: Optional[str]


class BlockFormat(str, Enum):
    """"""

    html = "html"
    md = "md"
    jupyter = "jupyter"
    tex = "tex"
    json = "json"
    bibtex = "bibtex"


class BlockChild(BaseModel):
    id: str
    src: BlockChildID


class BlockVersionLinks(SelfLinks):
    download: Optional[AnyHttpUrl]
    project: AnyHttpUrl
    block: Optional[AnyHttpUrl]
    versions: AnyHttpUrl
    drafts: Optional[AnyHttpUrl]
    created_by: AnyHttpUrl
    parent: Optional[AnyHttpUrl]


class ContentVersionPost(BaseModel):
    content: str
    format: Optional[BlockFormat]


class ArticleVersionPost(BaseModel):
    order: List[str]
    children: Dict[str, BlockChild]
    title: Optional[str]


class ImageVersionPost(BaseModel):
    upload_path: AnyHttpUrl


class NavigationItem(BaseModel):
    id: str
    parentId: Optional[str] = None
    title: str
    blockId: BlockID


class NavigationVersionPost(BaseModel):
    items: List[NavigationItem]


class Target(str, Enum):
    """"""

    JupyterMarkdown = "jupyter.markdown"
    JupyterRaw = "jupyter.raw"
    JupyterCode = "jupyter.code"
    JupyterOutput = "jupyter.output"


class OutputSummaryKind(str, Enum):
    """"""

    stream = "stream"
    text = "text"
    error = "error"
    image = "image"
    svg = "svg"
    html = "html"
    latex = "latex"
    json = "json"
    javascript = "javascript"
    plotly = "plotly"
    bokeh = "bokeh"
    ipywidgets = "ipywidgets"
    unknown = "unknown"


class CellOutputMimeTypes(str, Enum):
    """"""

    TextPlain = "text/plain"
    TextHtml = "text/html"
    TextLatex = "text/latex"
    ImagePng = "image/png"
    ImageBmp = "image/bmp"
    ImageJpeg = "image/jpeg"
    ImageSvg = "image/svg+xml"
    ImageGif = "image/gif"
    AppJson = "application/json"
    AppGeoJson = "application/geo+json"
    AppPlotly = "application/vnd.plotly.v1+json"
    AppVega = "application/vnd.vega.v5+json"
    AppVegaLite = "application/vnd.vegalite.v3+json"
    AppVirtualDom = "application/vdom.v1+json"
    AppJavascript = "application/javascript"
    AppWidgetView = "application/vnd.jupyter.widget-view+json"
    AppWidgetState = "application/vnd.jupyter.widget-state+json"
    AppBokehLoad = "application/vnd.bokehjs_load.v0+json"
    AppBokehExec = "application/vnd.bokehjs_exec.v0+json"


class OutputSummaryEntry(BaseModel):
    kind: OutputSummaryKind
    content_type: CellOutputMimeTypes
    content: Optional[str]
    link: Optional[str]


class OutputSummary(OutputSummaryEntry):
    alternate: Dict[OutputSummaryKind, OutputSummaryEntry]


# TODO becoming so optional, losing its usefulness? break out different models per kind?
class BlockVersion(BaseModel):
    id: BlockChildID
    kind: BlockKind
    format: Optional[BlockFormat]
    title: str
    description: str
    published: bool
    created_by: str
    date_created: datetime
    version: int
    parent: Optional[str]
    links: Optional[BlockVersionLinks]

    # Kind: Content
    targets: Optional[List[str]]
    content: Optional[str]  # references return objects
    metadata: Optional[dict]

    # Kind: Image
    size: Optional[int]
    content_type: Optional[str]
    md5: Optional[str]

    # Kind: Article
    date: Optional[datetime]
    order: Optional[List[str]]
    children: Optional[Dict[str, BlockChild]]

    # Kind: Navigation
    items: Optional[List[NavigationItem]]

    # Kind: Output
    targets: Optional[List[Target]]
    outputs: Optional[List[OutputSummary]]
    original: Optional[Any]
    upload_path: Optional[str]
    size: Optional[int]
    content_type: Optional[str]
    md5: Optional[str]


class BlockVersionList(BaseModel):
    items: List[BlockVersion]
    links: ListLinks


class BlockDraftData(BaseModel):
    title: str
    order: List[str]
    children: Dict[str, BlockChild]


class BlockDraftLinks(SelfLinks):
    block: AnyHttpUrl
    child: Optional[AnyHttpUrl]
    parent: Optional[AnyHttpUrl]


class BlockDraft(BaseModel):
    id: BlockChildID
    kind: BlockKind
    data: Union[BlockDraftData, BaseModel]
    parent: Optional[str]
    child: int
    locked: bool
    merged: bool
    merged_by: Optional[str]
    created_by: str
    date_created: datetime
    date_modified: datetime
    links: BlockDraftLinks


class BlockDraftList(BaseModel):
    items: List[BlockDraft]
    links: ListLinks


class AccessID(BaseModel):
    project: str
    user: str


class AccessKind(str, Enum):
    """"""

    user = "user"


class AccessRole(str, Enum):
    """"""

    project_owner = "project.owner"
    project_editor = "project.editor"
    project_comment = "project.comment"
    project_view = "project.view"


class AccessLinks(SelfLinks):
    project: AnyHttpUrl
    user: AnyHttpUrl


class Access(BaseModel):
    id: AccessID
    kind: AccessKind
    role: AccessRole
    date_created: datetime
    date_modified: datetime
    links: AccessLinks


class AccessList(BaseModel):
    items: List[Access]
    links: ListLinks


Item = TypeVar("Item")


class ItemList(GenericModel, Generic[Item]):
    items: List[Item]


class ChildrenList(BaseModel):
    blocks: ItemList[Block]
    versions: ItemList[BlockVersion]
    errors: List



import logging
import os
import re
import time
from collections import namedtuple
from typing import List

import requests

from ..client import API_URL, Session
from ..models import Block, BlockVersion, OutputSummaryKind

logger = logging.getLogger()

ImageFormats = {"image/png": "png", "image/jpeg": "jpg", "image/gif": "gif"}
ImageSummary = namedtuple("ImageSummary", ["content", "block_paths", "local_paths"])

LocalReferenceItem = namedtuple(
    "LocalReferenceItem", ["block_path", "local_tag", "bibtex"]
)

INLINE_IMAGE_BLOCK_REGEX = (
    r".*\\includegraphics.*{(block:[A-Za-z0-9]{20}/[A-Za-z0-9]{20}/[0-9]+)}.*\\*"
)
OUTPUT_IMAGE_BLOCK_REGEX = (
    r".*\\includegraphics.*{(block:[A-Za-z0-9]{20}/[A-Za-z0-9]{20}/[0-9]+"
    r"-output-[0-9]+)}.*\\*"
)
OUTPUT_SVG_BLOCK_REGEX = (
    r".*\\includesvg.*{(block:[A-Za-z0-9]{20}/[A-Za-z0-9]{20}/[0-9]+"
    r"-output-[0-9]+)}.*\\*"
)
INLINE_CITATION_BLOCK_REGEX = r"\\cite[pt]?{(block:[A-Za-z0-9]{20}/[A-Za-z0-9]{20})}"


def find_groups_in_content(regex_matcher: str, content: str):
    """
    Take a peice of tex content that we expect to contain one or more commands,
    find all instances based on the regex given and return the first groups

    Typically this is used to extract block_paths for different latex commands from
    TeX content
    """
    block_paths = []
    matches = re.finditer(regex_matcher, content)
    for match in matches:
        block_path = match[1]
        block_paths.append(block_path)

    return block_paths


def patch_local_images_from_tex_block(regex_matcher: str, content: str):
    """
    Take a peice of tex content that we expect to contain only one includegraphics directive
    get the remote path, assign a new local path and retrun a strcutre that can be parsed
    """
    updated_content = content
    block_paths = []
    local_paths = []
    matches = re.finditer(regex_matcher, content)
    for match in matches:
        block_path = match[1]
        hash_segment = block_path[6:]
        local_segment = hash_segment.replace("/", "_")
        local_path = block_path.replace("block:", "images/").replace(
            hash_segment, local_segment
        )
        updated_content = updated_content.replace(block_path, local_path)
        block_paths.append(block_path)
        local_paths.append(local_path)

    return ImageSummary(
        content=updated_content,
        block_paths=block_paths,
        local_paths=local_paths,
    )


def block_hash_to_url(block_hash: str):
    match = re.search(
        r"block:([A-Za-z0-9]{20})/([A-Za-z0-9]{20})/*([0-9]*)", block_hash
    )
    if match:
        if match[3]:
            return f"{API_URL}/blocks/{match[1]}/{match[2]}/versions/{match[3]}"
        return f"{API_URL}/blocks/{match[1]}/{match[2]}"
    raise ValueError(f"invalid block hash {block_hash}")


# TODO move to session - easier to mock
def get_model(session, url, model=BlockVersion):
    block = session._get_model(url, model)
    if not block:
        raise ValueError(f"Could not fetch the block {url}")
    return block


def fetch(url: str):
    resp = requests.get(url)
    if resp.status_code >= 400:
        raise ValueError(resp.content)
    return resp.content


def localize_images_from_content_block(
    session: Session, assets_folder: str, content: str
):
    patch = patch_local_images_from_tex_block(INLINE_IMAGE_BLOCK_REGEX, content)
    content_with_extensions = patch.content

    for block_path, local_path in zip(patch.block_paths, patch.local_paths):
        # get block
        url = block_hash_to_url(block_path)
        image_block = session._get_model(url, BlockVersion)
        if not image_block:
            raise ValueError(f"Could not fetch image block {url}")

        # download from links.download (signed) to local path
        if not image_block.links.download:
            raise ValueError(f"Block kind {image_block.kind} has no download link")

        resp = requests.get(image_block.links.download)
        if resp.status_code >= 400:
            raise ValueError(resp.content)

        # now that we have the block we need to update the local path in the content
        # and save the file with the correct file extension
        content_type = resp.headers.get("content-type")
        if content_type not in ImageFormats:
            raise ValueError(f"Unsupported image content type {content_type}")
        ext = ImageFormats[content_type]
        local_path_with_extension = f"{local_path}.{ext}"
        content_with_extensions = content_with_extensions.replace(
            local_path, local_path_with_extension
        )

        logging.info(f"Writing {local_path_with_extension}")
        with open(
            os.path.join(assets_folder, local_path_with_extension), "wb+"
        ) as file:
            file.write(resp.content)

    return content_with_extensions


def localize_images_from_output_block(assets_folder: str, output_version: BlockVersion):
    all_content = ""
    for i, output_summary in enumerate(output_version.outputs):
        if (
            output_summary.kind == OutputSummaryKind.image
            or output_summary.kind == OutputSummaryKind.svg
        ):
            logging.info("Output Summary %s" % output_summary.kind)
            patch = patch_local_images_from_tex_block(
                OUTPUT_IMAGE_BLOCK_REGEX
                if output_summary.kind == OutputSummaryKind.image
                else OUTPUT_SVG_BLOCK_REGEX,
                output_summary.content,
            )
            if patch is not None:
                # TODO won't work for non image outputs or multi outputs with
                # mixed image and non image
                if not output_summary.link:
                    raise ValueError(
                        f"OutputSummary {i} in version "
                        f"{output_version.id.project}/{output_version.id.block}/"
                        f"{output_version.id.version} has no download link"
                    )

                fetched_content = fetch(output_summary.link)

                extension = ImageFormats[output_summary.content_type.lower()]
                local_path_with_extension = f"{patch.local_paths[0]}.{extension}"
                patched_content_with_ext = patch.content.replace(
                    patch.local_paths[0], local_path_with_extension
                )
                all_content = all_content + "\n" + patched_content_with_ext

                logging.info(f"Writing {local_path_with_extension}")
                with open(
                    os.path.join(assets_folder, local_path_with_extension), "wb+"
                ) as file:
                    file.write(fetched_content)

        elif output_summary.kind == OutputSummaryKind.text:
            logging.info(f"Output Summary {output_summary.kind}")
            if output_summary.link:
                # text content can be large and may be stored in a file
                content = fetch(output_summary.link).decode("utf-8")
            else:
                content = output_summary.content
            all_content = all_content + "\n" + content
        else:
            logging.info(f"WARNING: Unknown Output Summary: {output_summary.kind}")
            patch = None

    return all_content


def localize_references_from_content_block(
    session: Session, reference_list: List[LocalReferenceItem], content: str
):
    """Looks for cite TeX commands in the content then replaces the block ids
    with locally unique identifiers based on the local reference list.

    The reference list is extended as new references are found (side effect)

    Appends a unique hash to each new reference encountered
    """
    block_paths = find_groups_in_content(INLINE_CITATION_BLOCK_REGEX, content)
    patched_content = content

    for block_path in block_paths:
        # check for the reference in the reference list based on the block_path
        matched_references = [r for r in reference_list if r.block_path == block_path]
        existing_reference = (
            matched_references[0] if (len(matched_references) > 0) else None
        )

        if existing_reference is None:
            url = block_hash_to_url(block_path)
            block = get_model(session, url, Block)

            # get latest version
            version = get_model(
                session, f"{url}/versions/{block.latest_version}?format=bibtex"
            )

            # update the list
            local_tag, plain_tag = parse_cite_tag_from_version(version.content)
            bibtex = version.content.replace(plain_tag, local_tag)

            reference_item = LocalReferenceItem(block_path, local_tag, bibtex)
            reference_list.append(reference_item)
            existing_reference = reference_item
            logging.info(f"using new reference {existing_reference.local_tag}")
        else:
            logging.info(f"using existing reference {existing_reference.local_tag}")

        # patch the content and move on
        patched_content = patched_content.replace(
            block_path, existing_reference.local_tag
        )

    return patched_content


def parse_cite_tag_from_version(content: str):
    tag = "ref"
    match = re.match("@article{([0-9a-zA-Z_]+),*\n", content)
    if match is not None:
        tag = match[1].replace(",", "")
    # adding quasi-random hash to save de-duping work
    hash_id = hex(int(time.time()))[2:]
    tag_with_hash = f"{tag}_{hash_id}"

    return tag_with_hash, tag
